local VERSION = tonumber(("$Revision: 46816 $"):match("%d+"))

if uab.revision < VERSION then
	uab.version = "r" .. VERSION
	uab.revision = VERSION
	uab.date = ("$Date: 2007-08-17 22:21:12 -0400 (Fri, 17 Aug 2007) $"):match("%d%d%d%d%-%d%d%-%d%d")
end

--[[------------------------------------------------------------------------------
The visual button for an instance of an action.  Based on settings, the button 
can look different.  It can look like a menu item, with the icon and spell name 
showing side by side, or look like a typical action button with just the icon 
showing.  Other appearances are possible as well.  This class controls 
implementing the standard appearance of the button, and how the visual effects 
actually appear (e.g., just the icon blinks, but the text does not, or the icon 
is highlighted, but the text is not).

An ActionButton instance is allocated for each action for a UnitGroupHeader.
Each ActionButton instance is associated with a single UnitGroupHeader and a 
single Action.

When an ActionButton is shown, it adjusts it's scale to the currently set scale,
then it tells the associated Action to identify the visual effects based on the 
state of the action and the unit the ActionButton is associated with.
--------------------------------------------------------------------------------]]
uabActionButton = uab:NewModule("ActionButton", "Metrognome-2.0", "AceEvent-2.0")

uabActionButton.core = uab
uabActionButton.minlevel = 0
uabActionButton.allbuttons = {}
uabActionButton.actionButtonFrames = {}
uabActionButton.visibleButtons = {}

local DogTag = AceLibrary("DogTag-1.0")

function uabActionButton:OnInitialize()
	uabActionButton.super.OnInitialize(self)
    
    --[[
    Start the repeated timed event to ensure that action button flashing
    occurs as expected.
    --]]
    self:RegisterMetro("ActionButtonTimer", self.ActionButtonTimer, 0, self)
    self:StartMetro("ActionButtonTimer")
    
    --[[
    Register for the standard game events that need to be watched for all
    buttons to function properly.
    --]]
    self:RegisterEvent("ACTIONBAR_UPDATE_USABLE", "UnitStatusChanged")          -- used to catch actions/items becoming usable/unusable
    self:RegisterEvent("ACTIONBAR_UPDATE_COOLDOWN", "UnitStatusChanged")        -- used to catch spell/action cooldowns
    self:RegisterEvent("PET_BAR_UPDATE_COOLDOWN", "UnitStatusChanged")          -- used to catch pet spell cooldowns
    self:RegisterEvent("BAG_UPDATE_COOLDOWN", "UnitStatusChanged")              -- used to catch item cooldowns
    self:RegisterEvent("PLAYER_AURAS_CHANGED", "UnitStatusChanged")             -- gain/lose a player buff/debuff (used to catch eating/drinking)
    self:RegisterEvent("PLAYER_ENTERING_WORLD", "UnitStatusChanged")            -- player enters the world, enters/leaves an instance, or respawns at a graveyard, or any loading screen
    self:RegisterEvent("PLAYER_REGEN_ENABLED", "UnitStatusChanged")             -- leave combat
    self:RegisterEvent("PLAYER_REGEN_DISABLED", "UnitStatusChanged")            -- enter combat
    self:RegisterEvent("PLAYER_ALIVE", "UnitStatusChanged")                     -- release to a graveyard or accepts a rez
    self:RegisterEvent("PLAYER_UNGHOST", "UnitStatusChanged")                   -- alive after being a ghost
    self:RegisterEvent("UnitActionBars_UnitStatus_Gained", "UnitStatusChanged") -- some unit status gained
    self:RegisterEvent("UnitActionBars_UnitStatus_Lost", "UnitStatusChanged")   -- some unit status lost

    self:RegisterEvent("UNIT_SPELLCAST_START", "StartCast")                     -- used to update checked state of last clicked action
    self:RegisterEvent("UNIT_SPELLCAST_STOP", "EndCast")                        -- used to update checked state of last clicked action
    self:RegisterEvent("UNIT_SPELLCAST_FAILED", "EndCast")                      -- used to update checked state of last clicked action
    self:RegisterEvent("UNIT_SPELLCAST_INTERRUPTED", "EndCast")                 -- used to update checked state of last clicked action
    self:RegisterEvent("UNIT_SPELLCAST_CHANNEL_START", "StartCast")             -- used to update checked state of last clicked action
    self:RegisterEvent("UNIT_SPELLCAST_CHANNEL_STOP", "EndCast")                -- used to update checked state of last clicked action
end

function uabActionButton:StartCast(unit)
    if unit == "player" then
        self:UnitStatusChanged()
    end
end

function uabActionButton:EndCast(unit)
    if unit == "player" then
        self:UnitStatusChanged()
    end
end

--[[
The function that is called when some event occurs that requires us to
update the visual characteristics of the action buttons.
--]]
function uabActionButton:UnitStatusChanged()
    for button, _ in pairs(self.visibleButtons) do
        local action = button:UnitStateModifiedAction()
        button:Update(action)
    end
end

--[[
The function that is called when the timed event occurs.  We call the
TimerEvent method of every visible action button to deal with the flashing
characteristics for that action button.
--]]
function uabActionButton:ActionButtonTimer(elapsed)
    for button, _ in pairs(self.visibleButtons) do
        local action = button:UnitStateModifiedAction()
        button:TimerEvent(elapsed, action)
    end
end

function uabActionButton:SetMinFrameLevel(minlevel)
    if minlevel > self.minlevel then
        self.minlevel = minlevel
        for button, _ in pairs(self.allbuttons) do
            button:SetFrameLevel(self.minlevel)
        end
    end
end

function uabActionButton:SetVisible(button)
    self.visibleButtons[button] = true
    button:Update()
end

function uabActionButton:ClearVisible(button)
    self.visibleButtons[button] = nil
end

function uabActionButton:GetActionButtons()
    return self.actionButtonFrames
end

function uabActionButton:RegisterActionButton(frame)
    self:TriggerEvent("UnitActionBars_ActionButtonCreated", frame)
    self.actionButtonFrames[frame] = true
end

function uabActionButton:New(header, action, number)
    local button = self.core.classes.ActionButton:new(header, action, number)
    self.allbuttons[button] = true
    return button
end

function uabActionButton:NewCancel(header)
    local button = self.core.classes.ActionButtonNone:new(header)
    self.allbuttons[button] = true
    return button
end

--[[------------------------------------------------------------------------------
Start of the instance class
--------------------------------------------------------------------------------]]
uab.classes.ActionButton = uab.mixins.aceOO.Class(uab.mixins.aceEvent)

function uab.classes.ActionButton.prototype:init(header, action, number)
	uab.classes.ActionButton.super.prototype.init(self)
	
	self.header = header
    self.posstates = {}
    self.showstates = {}
    self.savedattrs = {}
    self.flashing = {}
    self.visualeffect = {}
    
    self.number = number

    self.framename = "uab_"..self.header.name.."_ActionBtn"..self.number
    self.timername = "ActionButtonTimer_"..self.framename
    self.maxalpha = 1.0

    --[[
	Create the action button frame
	--]]
    self.frame = getglobal(self.framename)
    if not self.frame then
        self.frame = CreateFrame("CheckButton", self.framename, nil, "UAB_StandardActionButtonTemplate")
        self.frame.uabobj = self
        
        uabActionButton:RegisterActionButton(self.frame)
    end

    self.frame:SetFrameLevel(1)
    self.frame:SetFrameStrata("FULLSCREEN_DIALOG")
    self.frame:Hide()
    
    self.frame.icon = getglobal(self.frame:GetName().."Icon")
    self.frame.itemcount = getglobal(self.frame:GetName().."Count")
    self.frame.cooldown = getglobal(self.frame:GetName().."Cooldown")

--    self.frame.border = getglobal(self.frame:GetName().."Border")

    self.frame.border = self.frame:CreateTexture(self.frame:GetName().."Border", "OVERLAY")
    self.frame.border:SetBlendMode("ADD")
    self:SetBorderTexture("Interface\\AddOns\\UnitActionBars\\images\\swborder2")

    self.frame.normaltexture = getglobal(self.frame:GetName().."NormalTexture")

    -- border fix-ups to smooth icon appearance a bit (borrowed from Bongos)
	self.frame.icon:SetTexCoord(0.06, 0.94, 0.06, 0.94)
        
    self.frame.cooldown:SetFrameLevel(2)

    self.frame.uabobj = self
--    self.frame.cooldown:Show()
    self.frame.itemcount:Show()

    self.frame.border:Hide()

    self:HighlightBorderChanged()

    uabActionButton:SetBoolFrameAttribute(self.frame, "useparent-unit", true)
    uabActionButton:SetBoolFrameAttribute(self.frame, "useparent-unitsuffix", true)

    self:Enable()
    self:SetAction(action)

    self:SetParent()

    self.frame:SetScript("OnShow", function(self) self.uabobj:OnShow() end)
    self.frame:SetScript("PostClick", function(self, button) self.uabobj:PostClick(button) end)
    self.frame:SetScript("OnHide", function(self) self.uabobj:OnHide() end)
    self.frame:SetScript("OnEnter", function(self, motion) self.uabobj:ShowTooltip() end)
    self.frame:SetScript("OnLeave", function(self, motion) GameTooltip:Hide() end)

    --[[
    Start out with the button always hidden.  When some frame registers for the
    unit group of the activator for this button, the display properties of this
    button will get adjusted then.  We just need to start by indicating that the
    button is always hidden.
    --]]
    self:PrepareForLayouts()
    self:LayoutsDone()
end

function uab.classes.ActionButton.prototype:Dump(dumper)
    dumper:Dump("Action("..self.number.."): "..self.framename)
    dumper:IncIndent()
    if self.header then
        dumper:Dump("SecureHeader: "..self.header.frame:GetName())
    else
        dumper:Dump("SecureHeader: none")
    end
    
    dumper:DumpFrameAttribute(self.frame, "useparent-unit")
    dumper:DumpFrameAttribute(self.frame, "useparent-unitsuffix")
    dumper:DumpFrameAttribute(self.frame, "newstate")
    dumper:DumpFrameAttribute(self.frame, "ofspoint")
    dumper:DumpFrameAttribute(self.frame, "ofsx")
    dumper:DumpFrameAttribute(self.frame, "ofsy")
    dumper:DumpFrameAttribute(self.frame, "showstates")
    dumper:DumpFrameAttribute(self.frame, "hidestates")
    --self.action:Dump(dumper)
    dumper:DecIndent()
end

function uab.classes.ActionButton.prototype:PrepareForLayouts()
    for idx, _ in pairs(self.showstates) do
        self.showstates[idx] = nil
    end

    for x, _ in pairs(self.posstates) do
        self.posstates[x].pt = nil
        self.posstates[x].x = nil
        self.posstates[x].y = nil
        self.posstates[x] = nil
    end
end

function uab.classes.ActionButton.prototype:AddShowStates(states)
    self.showstates[states] = true
end

function uab.classes.ActionButton.prototype:SetPointForStates(states, ofspoint, ofsx, ofsy)
    self.posstates[states] = { pt = ofspoint, x = ofsx, y = ofsy }
end

function uab.classes.ActionButton.prototype:LayoutsDone()
    local ptattr, xattr, yattr
    
    for posstates, pos in pairs(self.posstates) do
        if not ptattr then
            ptattr = ""
            xattr = ""
            yattr = ""
        else
            ptattr = ptattr .. ";"
            xattr = xattr .. ";"
            yattr = yattr .. ";"
        end

        ptattr = ptattr .. posstates .. ":" .. pos.pt
        xattr = xattr .. posstates .. ":" .. pos.x
        yattr = yattr .. posstates .. ":" .. pos.y
    end

    uabActionButton:SetFrameAttribute(self.frame, "ofspoint", ptattr)
    uabActionButton:SetFrameAttribute(self.frame, "ofsx", xattr)
    uabActionButton:SetFrameAttribute(self.frame, "ofsy", yattr)

    local showstates

    for states, _ in pairs(self.showstates) do
        if not showstates then
            showstates = ""
        else
            showstates = showstates .. ","
        end

        showstates = showstates .. states
    end

    if not showstates then
        uabActionButton:SetFrameAttribute(self.frame, "showstates", nil)
        uabActionButton:SetFrameAttribute(self.frame, "hidestates", "*")
    else
        uabActionButton:SetFrameAttribute(self.frame, "showstates", showstates)
        uabActionButton:SetFrameAttribute(self.frame, "hidestates", nil)
    end
end

function uab.classes.ActionButton.prototype:SetParent()
    if self.header.frame then
        self.frame:SetParent(self.header.frame)
        self.frame:SetFrameLevel(1)
        self.frame:SetFrameStrata("FULLSCREEN_DIALOG")
        uabActionButton:SetFrameFrameAttribute(self.header.frame, "addchild", self.frame)
    end
end

function uab.classes.ActionButton.prototype:SetAction(action)
    if self.action then
        self.action:RemoveButton(self)
    end
    self.action = action
    self.actiondisabled = nil
    if self.action then
        self.action:AddButton(self)
    end
    
    self:CheckEnabledState()
end

function uab.classes.ActionButton.prototype:SetHostileAction(action)
    if self.hostileaction then
        self.hostileaction:RemoveButton(self)
    end
    self.hostileaction = action
    self.hostileactiondisabled = nil
    if self.hostileaction then
        self.hostileaction:AddButton(self)
    end
        
    self:CheckEnabledState()
end

function uab.classes.ActionButton.prototype:ClearAttributes(attributes)
	for name, _ in pairs(attributes) do
		if self.savedattrs[name] then
			uabActionButton:SetFrameAttribute(self.frame, name, self.savedattrs[name])
		else
			uabActionButton:SetFrameAttribute(self.frame, name, nil)
		end
	end
end

function uab.classes.ActionButton.prototype:SetAttributes(attributes)
	for name, val in pairs(attributes) do
		self.savedattrs[name] = self.frame:GetAttribute(name)
		uabActionButton:SetFrameAttribute(self.frame, name, val)
	end
end

function uab.classes.ActionButton.prototype:CurrentUnit()
    return uab.GetFrameUnit(self:Frame())
end

function uab.classes.ActionButton.prototype:UnitStateModifiedAction()
    local unit = self:CurrentUnit()
    local action = self.action
    
    if unit and UnitCanAttack("player", unit) and self.header.group.activator:HasHostileActions() then
        action = self.hostileaction
    end
    
    return action
end

function uab.classes.ActionButton.prototype:Show()
    local action = self:UnitStateModifiedAction()
                 
    if action and action:GetData() then
        self:StatusChanged(action)
    else
        self.frame:SetAlpha(0)
    end
end

function uab.classes.ActionButton.prototype:Hide()
	self.frame:Hide()
end

function uab.classes.ActionButton.prototype:DisableAction()
    self.actiondisabled = true
    self:CheckEnabledState()
end

function uab.classes.ActionButton.prototype:DisableHostileAction()
    self.hostileactiondisabled = true
    self:CheckEnabledState()
end

function uab.classes.ActionButton.prototype:ActionDisabled()
    if self.actiondisabled then
        return true
    end
    
    if not self.action or not self.action:GetData() then
        return true
    end
end

function uab.classes.ActionButton.prototype:HostileActionDisabled()
    if self.hostileactiondisabled then
        return true
    end
    
    if not self.hostileaction or not self.hostileaction:GetData() then
        return true
    end
end

function uab.classes.ActionButton.prototype:Enable()
    self.enabled = true
    self:CheckEnabledState()
end

function uab.classes.ActionButton.prototype:Disable()
    self.enabled = nil
    self:CheckEnabledState()
end

function uab.classes.ActionButton.prototype:CheckEnabledState()
    if self.enabled and not (self:ActionDisabled() and self:HostileActionDisabled()) then
        self:DoEnable()
    else
        self:DoDisable()
    end
end

function uab.classes.ActionButton.prototype:DoEnable()
    if not self.isenabled then
        self.isenabled = true
        self:StayOpenChanged()
    end
end

function uab.classes.ActionButton.prototype:DoDisable()
    if self.isenabled then
        self.isenabled = nil
        self:StayOpenChanged()    
    end
end

function uab.classes.ActionButton.prototype:StayOpenChanged()
    if not self.header.group:IsStayOpen() or not self.isenabled then
        uabActionButton:SetFrameAttribute(self.frame, "newstate", "*:0")
    else
        uabActionButton:SetFrameAttribute(self.frame, "newstate", nil)
    end
end

function uab.classes.ActionButton.prototype:HighlightBorderChanged()
    local width = uab.db.profile.borderwidth
     
    if width > 5 then
        width = 5
    elseif width < 2 then
        width = 2
    end

    self:SetBorderTexture("Interface\\AddOns\\UnitActionBars\\images\\swborder" .. width)
--    self.frame.border:SetTexture("Interface\\AddOns\\UnitActionBars\\images\\swborder" .. width)
end

function uab.classes.ActionButton.prototype:SetBorderTexture(texture)
    self.frame.border:SetTexture(texture)
    self.frame.border:SetWidth(40)
    self.frame.border:SetHeight(40)
    self.frame.border:ClearAllPoints()
    self.frame.border:SetPoint("CENTER", self.frame, "CENTER", 0, -1)
end

function uab.classes.ActionButton.prototype:Frame()
	return self.frame
end

function uab.classes.ActionButton.prototype:SetChecked(setit)
    if setit then
        self.frame:SetChecked(1)
    else
        self.frame:SetChecked(0)
    end
end

function uab.classes.ActionButton.prototype:BorderColor(r, g, b, a)
    if a == 0 then
        self.frame.border:Hide()
    else
        self.frame.border:SetVertexColor(r,g,b,a)
        self.frame.border:Show()
    end
end

function uab.classes.ActionButton.prototype:SetAlpha(alpha)
    self.maxalpha = alpha
end

function uab.classes.ActionButton.prototype:StartFlash()
    if not self.flashing.enabled then
        self.flashing.timer = uab.db.profile.blinkcycle/2
        self.flashing.direction = 1
        self.flashing.enabled = true
    end
end

function uab.classes.ActionButton.prototype:EndFlash()
    if self.flashing.enabled then
        self.flashing.enabled = nil
    end
end

function uab.classes.ActionButton.prototype:StatusChanged(action)
    if not action then
        action = self:UnitStateModifiedAction()
    end
    
    if action and action:GetData() then
        action:ShowButton(self)
    else
        self.frame:SetAlpha(0)
    end
end

function uab.classes.ActionButton.prototype:Update(action)
    if not action then
        action = self:UnitStateModifiedAction()
    end
    
    self:UpdateCooldown(action)
    self:StatusChanged(action)
end

function uab.classes.ActionButton.prototype:UpdateCooldown(action)
    if not action then
        action = self:UnitStateModifiedAction()
    end
    
    if action and action:GetData() then
        local start, duration, enable = action:GetCooldown(self)

        if enable and enable > 0 and start and start > 0 and duration and duration > 0 then
            self.oncooldown = true
            CooldownFrame_SetTimer(self.frame.cooldown, start, duration, enable)
        elseif self.oncooldown then
            self.oncooldown = nil
            CooldownFrame_SetTimer(self.frame.cooldown, 0, 0, 0)
        end
    elseif self.oncooldown then
        self.oncooldown = nil
        CooldownFrame_SetTimer(self.frame.cooldown, 0, 0, 0)
    end
end

function uab.classes.ActionButton.prototype:UpdateUsable(action)
    if not action then
        action = self:UnitStateModifiedAction()
    end
    
    if action and action:GetData() then
        action:UpdateVisualEffects(self)
        self:UpdateCooldown(action)
    else
        self.frame:SetAlpha(0)
    end
end

--[[------------------------------------------------------------------------------
Repeatedly called to cause/stop the flashing of a frame, when that is enabled
for an action button.
--------------------------------------------------------------------------------]]
function uab.classes.ActionButton.prototype:TimerEvent(elapsed, action)
    if not self.isenabled then
        return
    end
    
    local a = self.maxalpha
    if not action then
        action = self:UnitStateModifiedAction()
    end

    if action and action:GetData() then
        if self.flashing.enabled then
            self.flashing.timer = self.flashing.timer - elapsed
            if self.flashing.timer < 0 then
                self.flashing.timer = uab.db.profile.blinkcycle/2
                if self.flashing.direction then
                    self.flashing.direction = nil
                else
                    self.flashing.direction = 1
                end
            end
        
            if self.flashing.direction then
                a = ((self.flashing.timer * 2)/uab.db.profile.blinkcycle)*self.maxalpha
            else
                a = (1 - ((self.flashing.timer * 2)/uab.db.profile.blinkcycle))*self.maxalpha
            end
        end
    else
        a = 0
        self.flashing.enabled = nil
    end

    self.frame:SetAlpha(a)

    --Tooltip stuff, probably for the cooldown timer
    if self.updateTooltip then
        self.updateTooltip = self.updateTooltip - elapsed
        if self.updateTooltip <= 0 then
            if GameTooltip:IsOwned(self) then
                self:ShowTooltip(action)
            end
            self.updateTooltip = TOOLTIP_UPDATE_TIME
        end
    end
end

function uab.classes.ActionButton.prototype:IsUnit(unit)
    return unit == uab.GetFrameUnit(self.frame)
end

--[[------------------------------------------------------------------------------
Helper methods, to be used only by the class itself
--------------------------------------------------------------------------------]]
function uab.classes.ActionButton.prototype:ShowTooltip(action)
    if not action then
        action = self:UnitStateModifiedAction()
    end
             
    if action and action:GetData() then
        action:ShowTooltip(self.frame)
    end
end

function uab.classes.ActionButton.prototype:SetFrameLevel(level)
    self.frame:SetFrameLevel(level)
    self.frame.cooldown:SetFrameLevel(level+1)
    self.frame:SetFrameStrata("FULLSCREEN_DIALOG")
end

function uab.classes.ActionButton.prototype:OnShow()
    self.isvisible = true
    uabActionButton:SetVisible(self)
    
--[[
    -- parent starts as the header
    local parent = self.frame:GetParent()
    -- make parent be the control (parent of the header)
    parent = parent:GetParent()
    -- now, make it be the parent of the control (the unit frame)
    parent = parent:GetParent()

    self.frame:SetFrameLevel(parent:GetFrameLevel()+10)
--]]
    self.frame:SetFrameStrata("FULLSCREEN_DIALOG")
end

function uab.classes.ActionButton.prototype:PostClick(button)
    local msgInfo = self.action:GetMessage()
    if msgInfo then
        self:ShowActionMessage(msgInfo)
    end
end

function uab.classes.ActionButton.prototype:IsInRaid()
    return GetNumRaidMembers() ~= 0
end

function uab.classes.ActionButton.prototype:IsGrouped()
    return GetNumPartyMembers() ~= 0 or self:IsInRaid()
end

function uab.classes.ActionButton.prototype:ShowActionMessage(msgInfo)
    if msgInfo.enable and msgInfo.msg then
        local channel = nil
        local unit = uab.GetFrameUnit(self.frame)
        local msg

        if unit then
            msg = DogTag:Evaluate(unit, msgInfo.msg)
        
            if self:IsInRaid() then
                if msgInfo.groupch == "Raid" then
                    channel = "RAID"
                elseif msgInfo.groupch == "Party" then
                    channel = "PARTY"
                elseif msgInfo.groupch == "None" then
                    channel = "NONE"
                end
            end
            
            if not channel and self:IsGrouped() then
                if msgInfo.groupch == "Raid" or msgInfo.groupch == "Party" then
                    channel = "PARTY"
                elseif msgInfo.groupch == "None" then
                    channel = "NONE"
                end
            end
            
            if not channel then
                if msgInfo.defaultch == "Say" then
                    channel = "SAY"
                elseif msgInfo.defaultch == "Emote" then
                    channel = "EMOTE"
                elseif msgInfo.defaultch == "Channel" then
                    channel = "CHANNEL"
                elseif msgInfo.defaultch == "Self" then
                    channel = "SELF"
                else
                    channel = "NONE"
                end
            end
    
            if channel and channel ~= "NONE" then
                if channel == "SELF" then
                    uab.print(msg)
                elseif channel == "CHANNEL" then
                    SendChatMessage(msg, channel, "COMMON", msgInfo.channel)
                else
                    SendChatMessage(msg, channel)
                end
            end        
        end
    end
end

function uab.classes.ActionButton.prototype:OnHide()
    self.isvisible = nil
    uabActionButton:ClearVisible(self)
end

